"""Propagate a 1D wavefield using different implementations of an 8th order
finite difference method so that runtimes can be compared.
"""
import concurrent.futures
from ctypes import c_int, c_float
import numpy as np
from numba import jit
import wave_1d_fd_perf
from wave_1d_fd_perf import vfortran1
from wave_1d_fd_perf import vfortran2
from wave_1d_fd_perf import vfortran3

class Propagator(object):
    """An 8th order finite difference propagator for the 1D wave equation."""
    def __init__(self, model, dx, dt=None):
        self.nx = len(model)
        self.dx = np.float32(dx)
        max_vel = np.max(model)
        if dt:
            self.dt = dt
        else:
            self.dt = 0.6 * self.dx / max_vel
        self.nx_padded = self.nx + 2*8
        self.model_padded = np.pad(model, (8, 8), 'edge')
        self.model_padded2_dt2 = self.model_padded**2 * self.dt**2
        self.wavefield = [np.zeros(self.nx_padded, np.float32),
                          np.zeros(self.nx_padded, np.float32)
                         ]
        self.current_wavefield = self.wavefield[0]
        self.previous_wavefield = self.wavefield[1]

class VPy1(Propagator):
    """A simple Python implementation."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""
        for step in range(num_steps):
            f = self.current_wavefield
            fp = self.previous_wavefield

            for x in range(8, self.nx_padded-8):
                f_xx = (-735*f[x-8]+15360*f[x-7]
                        -156800*f[x-6]+1053696*f[x-5]
                        -5350800*f[x-4]+22830080*f[x-3]
                        -94174080*f[x-2]+538137600*f[x-1]
                        -924708642*f[x+0]
                        +538137600*f[x+1]-94174080*f[x+2]
                        +22830080*f[x+3]-5350800*f[x+4]
                        +1053696*f[x+5]-156800*f[x+6]
                        +15360*f[x+7]-735*f[x+8])/(302702400*self.dx**2)
                fp[x] = (self.model_padded[x]**2 * self.dt**2 * f_xx
                         + 2*f[x] - fp[x])

            for i in range(sources.shape[0]):
                sx = sources_x[i] + 8
                source_amp = sources[i, step]
                fp[sx] += (self.model_padded[sx]**2 * self.dt**2 * source_amp)

            self.current_wavefield = fp
            self.previous_wavefield = f

        return self.current_wavefield[8:self.nx_padded-8]


class VPy2(Propagator):
    """Like VPy1, but using model_padded2_dt2."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""
        for step in range(num_steps):
            f = self.current_wavefield
            fp = self.previous_wavefield

            for x in range(8, self.nx_padded-8):
                f_xx = (-735*f[x-8]+15360*f[x-7]
                        -156800*f[x-6]+1053696*f[x-5]
                        -5350800*f[x-4]+22830080*f[x-3]
                        -94174080*f[x-2]+538137600*f[x-1]
                        -924708642*f[x+0]
                        +538137600*f[x+1]-94174080*f[x+2]
                        +22830080*f[x+3]-5350800*f[x+4]
                        +1053696*f[x+5]-156800*f[x+6]
                        +15360*f[x+7]-735*f[x+8])/(302702400*self.dx**2)
                fp[x] = (self.model_padded2_dt2[x] * f_xx
                         + 2*f[x] - fp[x])

            for i in range(sources.shape[0]):
                sx = sources_x[i] + 8
                source_amp = sources[i, step]
                fp[sx] += (self.model_padded2_dt2[sx] * source_amp)

            self.current_wavefield = fp
            self.previous_wavefield = f

        return self.current_wavefield[8:self.nx_padded-8]


class VNumba1(Propagator):
    """A Numba implementation of inner loop."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        @jit(nopython=True)
        def _numba_step(f, fp, nx_padded, model_padded, dt, dx, sources,
                        sources_x):
            for x in range(8, nx_padded-8):
                f_xx = (-735*f[x-8]+15360*f[x-7]
                        -156800*f[x-6]+1053696*f[x-5]
                        -5350800*f[x-4]+22830080*f[x-3]
                        -94174080*f[x-2]+538137600*f[x-1]
                        -924708642*f[x+0]
                        +538137600*f[x+1]-94174080*f[x+2]
                        +22830080*f[x+3]-5350800*f[x+4]
                        +1053696*f[x+5]-156800*f[x+6]
                        +15360*f[x+7]-735*f[x+8])/(302702400*dx**2)
                fp[x] = (model_padded[x]**2 * dt**2 * f_xx
                         + 2*f[x] - fp[x])

            for i in range(sources.shape[0]):
                sx = sources_x[i] + 8
                fp[sx] += (model_padded[sx]**2 * dt**2 * sources[i])

        for step in range(num_steps):

            _numba_step(self.current_wavefield, self.previous_wavefield,
                        self.nx_padded, self.model_padded, self.dt,
                        self.dx, sources[:, step], sources_x)

            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VNumba2(Propagator):
    """A Numba implementation of outer loop."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        @jit(nopython=True)
        def _numba_step(f, fp, nx_padded, model_padded, dt, dx, sources,
                        sources_x, num_steps):

            for step in range(num_steps):
                for x in range(8, nx_padded-8):
                    f_xx = (-735*f[x-8]+15360*f[x-7]
                            -156800*f[x-6]+1053696*f[x-5]
                            -5350800*f[x-4]+22830080*f[x-3]
                            -94174080*f[x-2]+538137600*f[x-1]
                            -924708642*f[x+0]
                            +538137600*f[x+1]-94174080*f[x+2]
                            +22830080*f[x+3]-5350800*f[x+4]
                            +1053696*f[x+5]-156800*f[x+6]
                            +15360*f[x+7]-735*f[x+8])/(302702400*dx**2)
                    fp[x] = (model_padded[x]**2 * dt**2 * f_xx
                             + 2*f[x] - fp[x])

                for i in range(sources.shape[0]):
                    sx = sources_x[i] + 8
                    fp[sx] += (model_padded[sx]**2 * dt**2 * sources[i, step])

                tmp = f
                f = fp
                fp = tmp

            return f, fp

        self.current_wavefield, self.previous_wavefield = \
                _numba_step(self.current_wavefield, self.previous_wavefield,
                            self.nx_padded, self.model_padded, self.dt,
                            self.dx, sources, sources_x, num_steps)

        return self.current_wavefield[8:self.nx_padded-8]


class VNumba3(Propagator):
    """A parallel Numba implementation."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        @jit(nopython=True, nogil=True)
        def _numba_inner(f, fp, model_padded, dt, dx, start, end):
            for x in range(start, end):
                f_xx = (-735*f[x-8]+15360*f[x-7]
                        -156800*f[x-6]+1053696*f[x-5]
                        -5350800*f[x-4]+22830080*f[x-3]
                        -94174080*f[x-2]+538137600*f[x-1]
                        -924708642*f[x+0]
                        +538137600*f[x+1]-94174080*f[x+2]
                        +22830080*f[x+3]-5350800*f[x+4]
                        +1053696*f[x+5]-156800*f[x+6]
                        +15360*f[x+7]-735*f[x+8])/(302702400*dx**2)
                fp[x] = (model_padded[x]**2 * dt**2 * f_xx
                         + 2*f[x] - fp[x])

        def _numba_step(f, fp, model_padded, dt, dx, sources,
                        sources_x, num_steps, num_chunks=4):

            chunk_len = int(np.ceil((f.shape[0]-16) / num_chunks))
            chunks = np.zeros(num_chunks+1, np.int)
            chunks[:-1] = np.arange(8, f.shape[0]-8, chunk_len)
            chunks[-1] = f.shape[0]-8
            executor = concurrent.futures.ThreadPoolExecutor()

            for step in range(num_steps):

                fs = []

                for chunk_idx, chunk in enumerate(chunks[:-1]):
                    fs.append(executor.submit(_numba_inner, f, fp,
                                              model_padded, dt, dx,
                                              chunk, chunks[chunk_idx+1]))
                concurrent.futures.wait(fs)

                for i in range(sources.shape[0]):
                    sx = sources_x[i] + 8
                    fp[sx] += (model_padded[sx]**2 * dt**2 * sources[i, step])

                tmp = f
                f = fp
                fp = tmp

            return f, fp

        self.current_wavefield, self.previous_wavefield = \
                _numba_step(self.current_wavefield, self.previous_wavefield,
                            self.model_padded, self.dt,
                            self.dx, sources, sources_x, num_steps)

        return self.current_wavefield[8:self.nx_padded-8]


class VNumba4(Propagator):
    """Same as VNumba3, but using model_padded2_dt2."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        @jit(nopython=True, nogil=True)
        def _numba_inner(f, fp, model_padded2_dt2, dx, start, end):
            for x in range(start, end):
                f_xx = (-735*f[x-8]+15360*f[x-7]
                        -156800*f[x-6]+1053696*f[x-5]
                        -5350800*f[x-4]+22830080*f[x-3]
                        -94174080*f[x-2]+538137600*f[x-1]
                        -924708642*f[x+0]
                        +538137600*f[x+1]-94174080*f[x+2]
                        +22830080*f[x+3]-5350800*f[x+4]
                        +1053696*f[x+5]-156800*f[x+6]
                        +15360*f[x+7]-735*f[x+8])/(302702400*dx**2)
                fp[x] = (model_padded2_dt2[x] * f_xx
                         + 2*f[x] - fp[x])

        def _numba_step(f, fp, model_padded2_dt2, dx, sources,
                        sources_x, num_steps, num_chunks=4):

            chunk_len = int(np.ceil((f.shape[0]-16) / num_chunks))
            chunks = np.zeros(num_chunks+1, np.int)
            chunks[:-1] = np.arange(8, f.shape[0]-8, chunk_len)
            chunks[-1] = f.shape[0]-8
            executor = concurrent.futures.ThreadPoolExecutor()

            for step in range(num_steps):

                fs = []

                for chunk_idx, chunk in enumerate(chunks[:-1]):
                    fs.append(executor.submit(_numba_inner, f, fp,
                                              model_padded2_dt2, dx,
                                              chunk, chunks[chunk_idx+1]))
                concurrent.futures.wait(fs)

                for i in range(sources.shape[0]):
                    sx = sources_x[i] + 8
                    fp[sx] += (model_padded2_dt2[sx] * sources[i, step])

                tmp = f
                f = fp
                fp = tmp

            return f, fp

        self.current_wavefield, self.previous_wavefield = \
                _numba_step(self.current_wavefield, self.previous_wavefield,
                            self.model_padded2_dt2,
                            self.dx, sources, sources_x, num_steps)

        return self.current_wavefield[8:self.nx_padded-8]


class VFortran1(Propagator):
    """A Fortran implementation."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        vfortran1.vfortran1.step(self.current_wavefield,
                                 self.previous_wavefield,
                                 self.model_padded, self.dt, self.dx,
                                 sources, sources_x, num_steps,
                                 self.nx_padded, num_sources, source_len)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VFortran2(Propagator):
    """Like VFortran1, but using OpenMP on inner loop."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        vfortran2.vfortran2.step(self.current_wavefield,
                                 self.previous_wavefield,
                                 self.model_padded, self.dt, self.dx,
                                 sources, sources_x, num_steps,
                                 self.nx_padded, num_sources, source_len)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VFortran3(Propagator):
    """Like VFortran2, but with model_padded2_dt2."""
    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        vfortran3.vfortran3.step(self.current_wavefield,
                                 self.previous_wavefield,
                                 self.model_padded2_dt2, self.dx,
                                 sources, sources_x, num_steps,
                                 self.nx_padded, num_sources, source_len)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC1_gcc(Propagator):
    """A C implementation."""
    def __init__(self, model, dx, dt=None):
        super(VC1_gcc, self).__init__(model, dx, dt)

        self._libvc1 = np.ctypeslib.load_library('libvc1_gcc', wave_1d_fd_perf.__path__[0])
        self._libvc1.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float, c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc1.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded, self.dt, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC2_gcc_O2(Propagator):
    """Same as VC1, but using model_padded2_dt2."""
    def __init__(self, model, dx, dt=None):
        super(VC2_gcc_O2, self).__init__(model, dx, dt)
        self._libvc1 = np.ctypeslib.load_library('libvc2_gcc_O2', wave_1d_fd_perf.__path__[0])
        self._libvc1.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc1.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC2_gcc_O3(Propagator):
    """Same as VC2, but using -O3."""
    def __init__(self, model, dx, dt=None):
        super(VC2_gcc_O3, self).__init__(model, dx, dt)
        self._libvc1 = np.ctypeslib.load_library('libvc2_gcc_O3', wave_1d_fd_perf.__path__[0])
        self._libvc1.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc1.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC2_gcc_Ofast(Propagator):
    """Same as VC2, but using -Ofast."""
    def __init__(self, model, dx, dt=None):
        super(VC2_gcc_Ofast, self).__init__(model, dx, dt)
        self._libvc1 = np.ctypeslib.load_library('libvc2_gcc_Ofast', wave_1d_fd_perf.__path__[0])
        self._libvc1.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc1.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC2_gcc_Ofast_autopar(Propagator):
    """Same as VC2, but using -Ofast and autoparallelization."""
    def __init__(self, model, dx, dt=None):
        super(VC2_gcc_Ofast_autopar, self).__init__(model, dx, dt)
        self._libvc1 = np.ctypeslib.load_library('libvc2_gcc_Ofast_autopar',
                                                 wave_1d_fd_perf.__path__[0])
        self._libvc1.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc1.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC3_gcc(Propagator):
    """A C implementation with OpenMP of inner loop."""
    def __init__(self, model, dx, dt=None):
        super(VC3_gcc, self).__init__(model, dx, dt)
        self._libvc2 = np.ctypeslib.load_library('libvc3_gcc', wave_1d_fd_perf.__path__[0])
        self._libvc2.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float, c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc2.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded, self.dt, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC4_gcc(Propagator):
    """A C implementation with OpenMP of outer loop."""
    def __init__(self, model, dx, dt=None):
        super(VC4_gcc, self).__init__(model, dx, dt)
        self._libvc3 = np.ctypeslib.load_library('libvc4_gcc', wave_1d_fd_perf.__path__[0])
        self._libvc3.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float, c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc3.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded, self.dt, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC5_gcc_O2(Propagator):
    """Same as VC4 but using model_padded2_dt2."""
    def __init__(self, model, dx, dt=None):
        super(VC5_gcc_O2, self).__init__(model, dx, dt)
        self._libvc3 = np.ctypeslib.load_library('libvc5_gcc_O2', wave_1d_fd_perf.__path__[0])
        self._libvc3.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc3.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC5_gcc_O3(Propagator):
    """Same as VC5 but using -O3."""
    def __init__(self, model, dx, dt=None):
        super(VC5_gcc_O3, self).__init__(model, dx, dt)
        self._libvc3 = np.ctypeslib.load_library('libvc5_gcc_O3', wave_1d_fd_perf.__path__[0])
        self._libvc3.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc3.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]


class VC5_gcc_Ofast(Propagator):
    """Same as VC5 but using -Ofast."""
    def __init__(self, model, dx, dt=None):
        super(VC5_gcc_Ofast, self).__init__(model, dx, dt)
        self._libvc3 = np.ctypeslib.load_library('libvc5_gcc_Ofast', wave_1d_fd_perf.__path__[0])
        self._libvc3.step.argtypes = \
                [np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS', 'WRITEABLE')),
                 c_int,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                                        shape=(self.nx_padded),
                                        flags=('C_CONTIGUOUS')),
                 c_float,
                 np.ctypeslib.ndpointer(dtype=np.float32, ndim=2,
                                        flags=('C_CONTIGUOUS')),
                 np.ctypeslib.ndpointer(dtype=np.int, ndim=1,
                                        flags=('C_CONTIGUOUS')),
                 c_int, c_int, c_int]

    def step(self, num_steps, sources=None, sources_x=None):
        """Propagate wavefield."""

        num_sources = sources.shape[0]
        source_len = sources.shape[1]
        self._libvc3.step(self.current_wavefield, self.previous_wavefield,
                          self.nx_padded, self.model_padded2_dt2, self.dx,
                          sources, sources_x, num_sources, source_len,
                          num_steps)

        if num_steps%2 != 0:
            tmp = self.current_wavefield
            self.current_wavefield = self.previous_wavefield
            self.previous_wavefield = tmp

        return self.current_wavefield[8:self.nx_padded-8]

# loopy
# C: gcc, clang
# pylint



"""Test the propagators."""
import pytest
import numpy as np
from wave_1d_fd_perf.propagators import (VPy1, VPy2, VNumba1, VNumba2, VNumba3, VNumba4,
                             VFortran1, VFortran2, VFortran3,
                             VC1_gcc, VC2_gcc_O2, VC2_gcc_O3, VC2_gcc_Ofast,
                             VC2_gcc_Ofast_autopar, VC3_gcc, VC4_gcc,
                             VC5_gcc_O2, VC5_gcc_O3, VC5_gcc_Ofast)

def ricker(freq, length, dt, peak_time):
    """Return a Ricker wavelet with the specified central frequency."""
    t = np.arange(-peak_time, (length)*dt - peak_time, dt, dtype=np.float32)
    y = ((1.0 - 2.0*(np.pi**2)*(freq**2)*(t**2))
         * np.exp(-(np.pi**2)*(freq**2)*(t**2)))
    return y

def green(x0, x1, dx, dt, t, v, v0, f):
    """Use the 1D Green's function to determine the wavefield at a given
    location and time due to the given source.
    """
    y = np.sum(f[:np.maximum(0, int((t - np.abs(x1-x0)/v)/dt))])*dt*dx*v0/2
    return y

@pytest.fixture
def model_one():
    """Create a model with one reflector, and the expected wavefield."""
    N = 100
    rx = int(N/2)
    model = np.ones(N, dtype=np.float32) * 1500
    model[rx:] = 2500
    max_vel = 2500
    dx = 5
    dt = 0.001
    # 0.14 secs is chosen to avoid reflections from boundaries
    nsteps = np.ceil(0.14/dt).astype(np.int)
    source = ricker(25, nsteps, dt, 0.05)
    sx = 35
    expected = np.zeros(N)
    # create a new source shifted by the time to the reflector
    time_shift = np.round((rx-sx)*dx / 1500 / dt).astype(np.int)
    shifted_source = np.pad(source, (time_shift, 0), 'constant')
    # reflection and transmission coefficients
    r = (2500 - 1500) / (2500 + 1500)
    t = 1 + r

    # direct wave
    expected[:rx] = np.array([green(x*dx, sx*dx, dx, dt,
                                    (nsteps+1)*dt, 1500, 1500,
                                    source) for x in range(rx)])
    # reflected wave
    expected[:rx] += r*np.array([green(x*dx, (rx-1)*dx, dx, dt,
                                       (nsteps+1)*dt, 1500, 1500,
                                       shifted_source) for x in range(rx)])
    # transmitted wave
    expected[rx:] = t*np.array([green(x*dx, rx*dx, dx, dt,
                                      (nsteps+1)*dt, 2500, 1500,
                                      shifted_source) for x in range(rx, N)])
    return {'model': model, 'dx': dx, 'dt': dt, 'nsteps': nsteps,
            'sources': np.array([source]), 'sx': np.array([sx]),
            'expected': expected}

@pytest.fixture
def model_two():
    """Create a random model and compare with VPy1 implementation."""
    N = 100
    np.random.seed(0)
    model = np.random.random(N).astype(np.float32) * 3000 + 1500
    max_vel = 4500
    dx = 5
    dt = 0.6 * dx / max_vel
    nsteps = np.ceil(0.2/dt).astype(np.int)
    num_sources = 10
    sources_x = np.zeros(num_sources, dtype=np.int)
    sources = np.zeros([num_sources, nsteps], dtype=np.float32)
    for sourceIdx in range(num_sources):
        sources_x[sourceIdx] = np.random.randint(N)
        sources[sourceIdx, :] = ricker(25, nsteps, dt, 0.05+np.random.rand())
    v = VPy1(model, dx, dt)
    expected = v.step(nsteps, sources, sources_x)
    return {'model': model, 'dx': dx, 'dt': dt, 'nsteps': nsteps,
            'sources': sources, 'sx': sources_x, 'expected': expected}

@pytest.fixture
def versions():
    """Return a list of implementations."""
    return [VPy1, VPy2, VNumba1, VNumba2, VNumba3, VNumba4,
            VFortran1, VFortran2, VFortran3,
            VC1_gcc, VC2_gcc_O2, VC2_gcc_O3, VC2_gcc_Ofast,
            VC2_gcc_Ofast_autopar, VC3_gcc, VC4_gcc,
            VC5_gcc_O2, VC5_gcc_O3, VC5_gcc_Ofast]

def test_one_reflector(model_one, versions):
    """Verify that the numeric and analytic wavefields are similar."""

    for v in versions:
        _test_version(v, model_one, atol=1.5)


def test_allclose(model_two, versions):
    """Verify that all implementations produce similar results."""

    for v in versions[1:]:
        _test_version(v, model_two, atol=1e-8)


def _test_version(version, model, atol):
    """Run the test for one implementation."""
    v = version(model['model'], model['dx'], model['dt'])
    y = v.step(model['nsteps'], model['sources'], model['sx'])
    assert np.allclose(y, model['expected'], atol=atol)



from setuptools import setup, Extension
import numpy.distutils.core

vfortran1 = numpy.distutils.core.Extension(name='wave_1d_fd_perf.vfortran1', sources=['wave_1d_fd_perf/vfortran1.f90'])
vfortran2 = numpy.distutils.core.Extension(name='wave_1d_fd_perf.vfortran2', sources=['wave_1d_fd_perf/vfortran2.f90'], extra_compile_args=['-fopenmp'])
vfortran3 = numpy.distutils.core.Extension(name='wave_1d_fd_perf.vfortran3', sources=['wave_1d_fd_perf/vfortran3.f90'], extra_compile_args=['-fopenmp'])

libvc1_gcc = Extension(name='wave_1d_fd_perf.libvc1_gcc', sources=['wave_1d_fd_perf/vc1.c'], extra_compile_args=['-march=native', '-O3', '-std=c99'])

libvc2_gcc_O2 = Extension(name='wave_1d_fd_perf.libvc2_gcc_O2', sources=['wave_1d_fd_perf/vc2.c'], extra_compile_args=['-march=native', '-O2', '-std=c99'])

libvc2_gcc_O3 = Extension(name='wave_1d_fd_perf.libvc2_gcc_O3', sources=['wave_1d_fd_perf/vc2.c'], extra_compile_args=['-march=native', '-O3', '-std=c99'])
	
libvc2_gcc_Ofast = Extension(name='wave_1d_fd_perf.libvc2_gcc_Ofast', sources=['wave_1d_fd_perf/vc2.c'], extra_compile_args=['-march=native', '-Ofast', '-std=c99'])

libvc2_gcc_Ofast_autopar = Extension(name='wave_1d_fd_perf.libvc2_gcc_Ofast_autopar', sources=['wave_1d_fd_perf/vc2.c'], extra_compile_args=['-march=native', '-Ofast', '-floop-parallelize-all', '-ftree-parallelize-loops=4', '-std=c99'])
	
libvc3_gcc = Extension(name='wave_1d_fd_perf.libvc3_gcc', sources=['wave_1d_fd_perf/vc3.c'], extra_compile_args=['-march=native', '-O3', '-fopenmp', '-std=c99'])

libvc4_gcc = Extension(name='wave_1d_fd_perf.libvc4_gcc', sources=['wave_1d_fd_perf/vc4.c'], extra_compile_args=['-march=native', '-O3', '-fopenmp', '-std=c99'])

libvc5_gcc_O2 = Extension(name='wave_1d_fd_perf.libvc5_gcc_O2', sources=['wave_1d_fd_perf/vc5.c'], extra_compile_args=['-march=native', '-O2', '-fopenmp', '-std=c99'])

libvc5_gcc_O3 = Extension(name='wave_1d_fd_perf.libvc5_gcc_O3', sources=['wave_1d_fd_perf/vc5.c'], extra_compile_args=['-march=native', '-O3', '-fopenmp', '-std=c99'])

libvc5_gcc_Ofast = Extension(name='wave_1d_fd_perf.libvc5_gcc_Ofast', sources=['wave_1d_fd_perf/vc5.c'], extra_compile_args=['-march=native', '-Ofast', '-fopenmp', '-std=c99'])

numpy.distutils.core.setup(
        name='wave_1d_fd_perf',
        version='0.0.11',
        description='Performance analysis of different implementations of 1d finite difference wave propagation',
        url='https://github.com/ar4/wave_1d_fd_perf',
        author='Alan Richardson',
        author_email='alan@ausargeo.com',
        license='MIT',
        packages=['wave_1d_fd_perf'],
        install_requires=['numpy','pandas','numba'],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
        ],
        ext_modules=[vfortran1, vfortran2, vfortran3, libvc1_gcc, libvc2_gcc_O2, libvc2_gcc_O3, libvc2_gcc_Ofast, libvc2_gcc_Ofast_autopar, libvc3_gcc, libvc4_gcc, libvc5_gcc_O2, libvc5_gcc_O3, libvc5_gcc_Ofast]
)



"""Measure the runtime of the propagators."""
from timeit import repeat
import numpy as np
import pandas as pd
from wave_1d_fd_perf.propagators import (VPy1, VPy2, VNumba1, VNumba2, VNumba3, VNumba4,
                             VFortran1, VFortran2, VFortran3,
                             VC1_gcc, VC2_gcc_O2, VC2_gcc_O3, VC2_gcc_Ofast,
                             VC2_gcc_Ofast_autopar, VC3_gcc, VC4_gcc,
                             VC5_gcc_O2, VC5_gcc_O3, VC5_gcc_Ofast)
from wave_1d_fd_perf.test_wave_1d_fd_perf import ricker

def run_timing_num_steps(num_repeat=10, num_steps=range(0, 110, 10), model_size=1000):
    """Time implementations as num_steps varies."""

    versions = _versions()

    times = pd.DataFrame(columns=['version', 'num_steps', 'model_size', 'time'])

    for nsteps in num_steps:
        model = _make_model(model_size, nsteps)
        times = _time_versions(versions, model, num_repeat, times)

    return times


def run_timing_model_size(num_repeat=10, num_steps=20, model_sizes=range(200, 2200, 200)):
    """Time implementations as model size varies."""

    versions = _versions()

    times = pd.DataFrame(columns=['version', 'num_steps', 'model_size', 'time'])

    for N in model_sizes:
        model = _make_model(N, num_steps)
        times = _time_versions(versions, model, num_repeat, times)

    return times


def _versions():
    """Return a list of versions to be timed."""
    return [{'class': VPy1, 'name': 'Python v1'},
            {'class': VPy2, 'name': 'Python v2'},
            {'class': VNumba1, 'name': 'Numba v1'},
            {'class': VNumba2, 'name': 'Numba v2'},
            {'class': VNumba3, 'name': 'Numba v3'},
            {'class': VNumba4, 'name': 'Numba v4'},
            {'class': VFortran1, 'name': 'Fortran v1'},
            {'class': VFortran2, 'name': 'Fortran v2'},
            {'class': VFortran3, 'name': 'Fortran v3'},
            {'class': VC1_gcc, 'name': 'C v1 (gcc, -O3)'},
            {'class': VC2_gcc_O2, 'name': 'C v2 (gcc, -O2)'},
            {'class': VC2_gcc_O3, 'name': 'C v2 (gcc, -O3)'},
            {'class': VC2_gcc_Ofast, 'name': 'C v2 (gcc, -Ofast)'},
            {'class': VC2_gcc_Ofast_autopar, 'name': 'C v2 (gcc, -Ofast, autoparallel)'},
            {'class': VC3_gcc, 'name': 'C v3 (gcc, -O3)'},
            {'class': VC4_gcc, 'name': 'C v4 (gcc, -O3)'},
            {'class': VC5_gcc_O2, 'name': 'C v5 (gcc, -O2)'},
            {'class': VC5_gcc_O3, 'name': 'C v5 (gcc, -O3)'},
            {'class': VC5_gcc_Ofast, 'name': 'C v5 (gcc, -Ofast)'}]


def _make_model(N, nsteps):
    """Create a model with a given number of elements and time steps."""
    model = np.random.random(N).astype(np.float32) * 3000 + 1500
    max_vel = 4500
    dx = 5
    dt = 0.6 * dx / max_vel
    source = ricker(25, nsteps, dt, 0.05)
    sx = int(N/2)
    return {'model': model, 'dx': dx, 'dt': dt, 'nsteps': nsteps,
            'sources': np.array([source]), 'sx': np.array([sx])}


def _time_versions(versions, model, num_repeat, dataframe):
    """Loop over versions and append the timing results to the dataframe."""
    num_steps = model['nsteps']
    model_size = len(model['model'])
    for v in versions:
        time = _time_version(v['class'], model, num_repeat)
        dataframe = dataframe.append({'version': v['name'],
                                      'num_steps': num_steps,
                                      'model_size': model_size,
                                      'time': time}, ignore_index=True)
    return dataframe


def _time_version(version, model, num_repeat):
    """Time a particular version."""
    v = version(model['model'], model['dx'], model['dt'])

    def closure():
        """Closure over variables so they can be used in repeat below."""
        v.step(model['nsteps'], model['sources'], model['sx'])
 
    return np.min(repeat(closure, number=1, repeat=num_repeat))

if __name__ == '__main__':
    print(run_timing_num_steps())
